from .monitored_asyncio import (
    MonitoredAsyncIOEventLoopPolicy,
    monitored_asyncio_loop_factory,
)
from .monitored_uvloop import (
    MonitoredUvloopEventLoopPolicy,
    monitored_uvloop_loop_factory,
)

__all__ = [
    "MonitoredAsyncIOEventLoopPolicy",
    "MonitoredUvloopEventLoopPolicy",
    "monitored_asyncio_loop_factory",
    "monitored_uvloop_loop_factory",
]



import typing
from asyncio.events import BaseDefaultEventLoopPolicy
from monitored_ioloop.monitoring import IoLoopMonitorState


class BaseMonitoredEventLoopPolicy(BaseDefaultEventLoopPolicy):
    def __init__(
        self,
        monitor_callback: typing.Callable[[IoLoopMonitorState], None],
    ):
        super().__init__()
        self._monitor_callback = monitor_callback

    if typing.TYPE_CHECKING:
        # EventLoopPolicy doesn't implement these, but since they are marked
        # as abstract in typeshed, we have to put them in so mypy thinks
        # the base methods are overridden. This is the same approach taken
        # for the Windows event loop policy classes in typeshed.
        # pragma: no cover
        def get_child_watcher(self) -> typing.NoReturn: ...

        # pragma: no cover
        def set_child_watcher(self, watcher: typing.Any) -> typing.NoReturn: ...



class NoUvLoopInstalled(ImportError):
    def __init__(self) -> None:
        super().__init__(
            "Please install uvloop compatible version via `pip install monitored_ioloop[uvloop]`"
        )



import asyncio

from starlette.types import ASGIApp, Receive, Scope, Send
from typing import Protocol


class AsgiMiddlewareType(Protocol):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None: ...


def mask_numeric_segments(path: str) -> str:
    """
    Example:
    mask_numeric_segments("/api/v1/users/123/name") -> "/api/v1/users/_/name"
    """
    path_parts = path.split("/")
    return "/".join((part if not part.isdigit() else "_") for part in path_parts)


def default_callback_pretty_name(scope: Scope) -> str:
    """
    Default callback_pretty_name for FastAPI helper middleware.
    """
    masked_path = mask_numeric_segments(scope["path"])
    return f"[{scope['method']}] {masked_path}"


class MonitoredIOLoopMiddleware:
    """
    This feature requires you to have installed fastapi BY YOURSELF.
    monitored_ioloop DOES NOT REQUIRE fastapi as a dependency.
    In order to allow a more useful callback_pretty_name when using starlette based frameworks (for example FastAPI),
    this middleware will set the current task name to the HTTP method and path.
    For example when a GET request is made to /ping, the current task name will be set to "[GET] /ping".
    """

    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        current_task = asyncio.current_task()
        if current_task is not None:
            current_task.set_name(default_callback_pretty_name(scope))

        await self.app(scope, receive, send)



import asyncio
import typing
import warnings
from asyncio import Handle

from mypy_extensions import VarArg

from monitored_ioloop.monitored_ioloop_base import BaseMonitoredEventLoopPolicy
from monitored_ioloop.monitoring import (
    wrap_callback_with_monitoring,
    IoLoopMonitorState,
    IoLoopInnerState,
)


_Ts = typing.TypeVarTuple("_Ts")


class MonitoredSelectorEventLoop(asyncio.SelectorEventLoop):
    def __init__(
        self,
        monitor_callback: typing.Callable[[IoLoopMonitorState], None],
        *args: typing.Any,
        **kwargs: typing.Any,
    ):
        super().__init__(*args, **kwargs)
        self._monitor_callback = monitor_callback
        self._state = IoLoopInnerState(handles_count=0)

    def call_soon(
        self,
        callback: typing.Callable[[VarArg(*_Ts)], object],
        *args: *_Ts,
        **kwargs: typing.Any,
    ) -> Handle:
        callback_with_monitoring = wrap_callback_with_monitoring(
            callback, self._monitor_callback, self._state
        )

        handle = super().call_soon(callback_with_monitoring, *args, **kwargs)
        callback_with_monitoring.set_handle(handle)
        return handle


class MonitoredAsyncIOEventLoopPolicy(BaseMonitoredEventLoopPolicy):
    """Event loop policy.

    .. deprecated:: 0.0.15
        Use :func:`monitored_asyncio_loop_factory` instead.
        The policy-based approach is deprecated in favor of the loop factory approach
        which is preferred by Python's asyncio documentation.

    Usage example:
    >>> import asyncio
    >>> import monitored_ioloop
    >>> asyncio.set_event_loop_policy(monitored_ioloop.MonitoredAsyncIOEventLoopPolicy())
    >>> asyncio.get_event_loop()
    """

    def __init__(self, monitor_callback: typing.Callable[[IoLoopMonitorState], None]):
        warnings.warn(
            "MonitoredAsyncIOEventLoopPolicy is deprecated. "
            "Use monitored_uvloop_loop_factory() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(monitor_callback)

    def _loop_factory(self) -> MonitoredSelectorEventLoop:
        loop = MonitoredSelectorEventLoop(self._monitor_callback)
        return loop


def monitored_asyncio_loop_factory(
    monitor_callback: typing.Callable[[IoLoopMonitorState], None],
) -> typing.Callable[[], MonitoredSelectorEventLoop]:
    """Create a loop factory function for use with asyncio.run().

    Usage:
    >>> import asyncio
    >>> import monitored_ioloop
    >>> factory = monitored_ioloop.monitored_asyncio_loop_factory(lambda state: print(state))
    >>> asyncio.run(main(), loop_factory=factory)
    """

    def factory() -> MonitoredSelectorEventLoop:
        return MonitoredSelectorEventLoop(monitor_callback)

    return factory



import threading
import time
import typing
from asyncio import Handle
from dataclasses import dataclass
from logging import getLogger

import wrapt

from monitored_ioloop.formatting_utils import pretty_format_handle, pretty_callback_name

logger = getLogger(__name__)


@dataclass
class IoLoopMonitorState:
    """
    A dataclass containing the state of the loop when the callback was executed.
    This class is the interface that the monitor callback will receive.

    A basic Lexicon:
    * Handle - A wrapper for a callback that is scheduled to be executed by the loop.
    * Callback - The function that is executed by the loop.
    * Loop - The event loop that is executing the callbacks.
    """

    """
    Wall executing time of the callback
    It can be the whole coroutine or parts of it, depending on if the executing control
    was delegated back the loop or not.

    Wall Time explanation - https://en.wikipedia.org/wiki/Wall-clock_time
    """
    callback_wall_time: float

    """
    A best effort try to give a meaningful name to the callback that was currently executed.
    This property will come in handy when trying to debug callbacks with high wall time. 
    """
    callback_pretty_name: str

    """
    The amount of handles in the loop, excluding the current one.
    """
    loop_handles_count: int

    """
    The amount of time it took from the moment the coroutine was added to the loop until it was executed.
    """
    loop_lag: float


@dataclass
class IoLoopInnerState:
    handles_count: int
    handles_count_lock = threading.Lock()

    def increase_handles_count_thread_safe(self, increase_by: int) -> None:
        """
        Increase the amount of total handles in a thread safe way.
        """
        with self.handles_count_lock:
            self.handles_count += increase_by

    def decrease_handles_count_thread_safe(self, decrease_by: int) -> None:
        """
        Decrease the amount of total handles in a thread safe way.
        """
        with self.handles_count_lock:
            self.handles_count -= decrease_by


class MonitoredCallbackWrapper(wrapt.ObjectProxy):  # type: ignore
    def __init__(
        self,
        callback: typing.Callable[..., typing.Any],
        monitor_callback: typing.Callable[[IoLoopMonitorState], None],
        io_loop_state: IoLoopInnerState,
    ):
        super().__init__(callback)
        self._self_original_callback = callback
        self._self_monitor_callback = monitor_callback
        self._self_ioloop_state = io_loop_state
        self._self_added_to_loop_time = time.perf_counter()
        self._self_handle: typing.Optional[Handle] = None

    def set_handle(self, handle: Handle) -> None:
        self._self_handle = handle

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        loop_lag = time.perf_counter() - self._self_added_to_loop_time
        start_wall_time = time.perf_counter()
        response = self._self_original_callback(*args, **kwargs)
        self._self_ioloop_state.decrease_handles_count_thread_safe(1)
        wall_duration = time.perf_counter() - start_wall_time

        try:
            pretty_name = (
                pretty_format_handle(self._self_handle)
                if self._self_handle
                else pretty_callback_name(self._self_original_callback)
            )
            self._self_monitor_callback(
                IoLoopMonitorState(
                    callback_wall_time=wall_duration,
                    loop_handles_count=self._self_ioloop_state.handles_count,
                    loop_lag=loop_lag,
                    callback_pretty_name=pretty_name,
                )
            )
        except Exception:
            logger.warning("Monitor callback failed.", exc_info=True)
        return response


def wrap_callback_with_monitoring(
    callback: typing.Callable[..., typing.Any],
    monitor_callback: typing.Callable[[IoLoopMonitorState], None],
    ioloop_state: IoLoopInnerState,
) -> MonitoredCallbackWrapper:
    ioloop_state.increase_handles_count_thread_safe(1)
    return MonitoredCallbackWrapper(callback, monitor_callback, ioloop_state)



import typing
from asyncio import tasks
from asyncio.events import Handle


def pretty_format_handle(handle: Handle) -> str:
    callback = handle._callback  # type: ignore
    if isinstance(getattr(callback, "__self__", None), tasks.Task):
        # format the task
        return repr(callback.__self__)
    else:
        return repr(handle)


def pretty_callback_name(callback: typing.Callable[..., typing.Any]) -> str:
    if isinstance(getattr(callback, "__self__", None), tasks.Task):
        # format the task
        return getattr(
            callback.__self__,  # type: ignore
            "__qualname__",
            None,
        ) or repr(
            callback.__self__  # type: ignore
        )
    else:
        return getattr(callback, "__qualname__", None) or repr(callback)
